==> Smdn.Net.Imap4.Client-1.04.tar.bz2: Smdn.Net.Imap4.Client-1.04/Smdn.Net.Imap4.Client/Test/Smdn.Net.Imap4.Client/TestUtils.cs <==
using System;
using System.Linq;
using NUnit.Framework;

using Smdn.Net.Imap4.Client.Session;

namespace Smdn.Net.Imap4.Client {
  public static class TestUtils {
    public static void TestAuthenticated(Action<ImapPseudoServer, ImapClient> testAction)
    {
      TestAuthenticated(new ImapCapability[0], testAction);
    }

    public static void TestAuthenticated(ImapCapability[] serverCapabilities, Action<ImapPseudoServer, ImapClient> testAction)
    {
      var capas = new ImapCapabilitySet(serverCapabilities ?? new ImapCapability[0]);

      try {
        capas.Add(ImapCapability.Imap4Rev1);
      }
      catch {
        // ignore exception
      }

      using (var server = new ImapPseudoServer()) {
        server.Start();

        // greeting
        server.EnqueueResponse("* OK ImapPseudoServer ready\r\n");
        // CAPABILITY
        server.EnqueueTaggedResponse("* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");
        // LOGIN
        server.EnqueueTaggedResponse("$tag OK done\r\n");
        // CAPABILITY
        server.EnqueueTaggedResponse(string.Format("* CAPABILITY {0}\r\n", string.Join(" ", capas.ToStringArray())) +
                                     "$tag OK done\r\n");

        using (var client = new ImapClient(new Uri(string.Format("imap://user@{0}/", server.HostPort)))) {
          client.Timeout = 5000;
          client.Profile.AllowInsecureLogin = true;
          client.Profile.UsingSaslMechanisms = null;
          client.Profile.UseTlsIfAvailable = false;

          client.Connect("pass");

          StringAssert.EndsWith("CAPABILITY\r\n", server.DequeueRequest());
          StringAssert.EndsWith("LOGIN user pass\r\n", server.DequeueRequest());
          StringAssert.EndsWith("CAPABILITY\r\n", server.DequeueRequest());

          testAction(server, client);
        }
      }
    }

    public static void TestOpenedMailbox(string mailboxName, Action<ImapPseudoServer, ImapOpenedMailboxInfo> action)
    {
      TestOpenedMailbox(null, mailboxName, "$tag OK [READ-WRITE] done\r\n", action);
    }

    public static void TestOpenedMailbox(string mailboxName, string selectResponse, Action<ImapPseudoServer, ImapOpenedMailboxInfo> action)
    {
      TestOpenedMailbox(null, mailboxName, selectResponse, action);
    }

    public static void TestOpenedMailbox(ImapCapability[] capabilities, string mailboxName, string selectResponse, Action<ImapPseudoServer, ImapOpenedMailboxInfo> action)
    {
      TestUtils.TestAuthenticated(capabilities, delegate(ImapPseudoServer server, ImapClient client) {
        // LIST
        server.EnqueueTaggedResponse(string.Format("* LIST () \"\" {0}\r\n", mailboxName) +
                                     "$tag OK done\r\n");
        // SELECT
        server.EnqueueTaggedResponse(selectResponse);

        using (var opened = client.OpenMailbox(mailboxName)) {
          server.DequeueRequest(); // LIST
          server.DequeueRequest(); // SELECT

          try {
            Assert.IsTrue(opened.IsOpen);
            Assert.AreEqual(opened.IsReadOnly, selectResponse.Contains("OK [READ-ONLY]"));
            Assert.AreEqual(mailboxName, opened.FullName);
            Assert.IsNotNull(opened.ApplicableFlags);
            Assert.IsNotNull(opened.PermanentFlags);

            action(server, opened);
          }
          finally {
            // CLOSE
            if (client.IsConnected && opened.IsOpen)
              server.EnqueueTaggedResponse("$tag OK done\r\n");
          }
        }
      });
    }

    public static void TestClosedMailbox(string mailboxName, Action<ImapPseudoServer, ImapOpenedMailboxInfo> action)
    {
      TestOpenedMailbox(mailboxName, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        // CLOSE
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        mailbox.Close();

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsNull(mailbox.Client.OpenedMailbox);

        Assert.That(server.DequeueRequest(), Is.StringEnding("CLOSE\r\n"));

        action(server, mailbox);
      });
    }
  }

  public static class TestUtils2 {
    public static void TestAuthenticated(Action<ImapPseudoServer2, ImapClient> testAction)
    {
      TestAuthenticated(new ImapCapability[0],
                        testAction);
    }

    public static void TestAuthenticated(ImapCapability[] serverCapabilities,
                                         Action<ImapPseudoServer2, ImapClient> testAction)
    {
      var capas = new ImapCapabilitySet(serverCapabilities ?? new ImapCapability[0]);

      try {
        capas.Add(ImapCapability.Imap4Rev1);
      }
      catch {
        // ignore exception
      }

      using (var server = ImapPseudoServer2.StartNew()) {
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     string.Format("* CAPABILITY {0}\r\n", string.Join(" ", capas.ToStringArray())) +
                                     "$tag OK done\r\n");

        using (var client = new ImapClient(new Uri(string.Format("imap://user@{0}/", server.HostPort)))) {
          client.Timeout = 5000;
          client.Profile.AllowInsecureLogin = true;
          client.Profile.UsingSaslMechanisms = null;
          client.Profile.UseTlsIfAvailable = false;

          client.Connect("pass");

          server.DequeueRequest(); // CAPABILITY
          server.DequeueRequest(); // LOGIN
          server.DequeueRequest(); // CAPABILITY

          testAction(server, client);
        }
      }
    }

    public static void TestOpenedMailbox(string mailboxName,
                                         Action<ImapPseudoServer2, ImapOpenedMailboxInfo> action)
    {
      TestOpenedMailbox(null,
                        mailboxName,
                        "$tag OK [READ-WRITE] done\r\n",
                        action);
    }

    public static void TestOpenedMailbox(string mailboxName,
                                         string selectResponse,
                                         Action<ImapPseudoServer2, ImapOpenedMailboxInfo> action)
    {
      TestOpenedMailbox(null,
                        mailboxName,
                        selectResponse,
                        action);
    }

    public static void TestOpenedMailbox(ImapCapability[] capabilities,
                                         string mailboxName,
                                         string selectResponse,
                                         Action<ImapPseudoServer2, ImapOpenedMailboxInfo> action)
    {
      TestAuthenticated(capabilities,
                        delegate(ImapPseudoServer2 server, ImapClient client) {
        server.EnqueueTaggedResponse("LIST \"\"",
                                     string.Format("* LIST () \"\" {0}\r\n", mailboxName) +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("SELECT " + mailboxName,
                                     selectResponse);

        using (var opened = client.OpenMailbox(mailboxName)) {
          server.DequeueRequest(); // LIST
          server.DequeueRequest(); // SELECT

          try {
            Assert.IsTrue(opened.IsOpen);
            Assert.AreEqual(opened.IsReadOnly, selectResponse.Contains("OK [READ-ONLY]"));
            Assert.AreEqual(mailboxName, opened.FullName);
            Assert.IsNotNull(opened.ApplicableFlags);
            Assert.IsNotNull(opened.PermanentFlags);

            action(server, opened);
          }
          finally {
            // CLOSE
            if (client.IsConnected && opened.IsOpen)
              server.EnqueueTaggedResponse("CLOSE\r\n",
                                           "$tag OK done\r\n");

          }
        }
      });
    }

    public static void TestClosedMailbox(string mailboxName,
                                         Action<ImapPseudoServer2, ImapOpenedMailboxInfo> action)
    {
      TestOpenedMailbox(mailboxName,
                        delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK done\r\n");

        mailbox.Close();

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsNull(mailbox.Client.OpenedMailbox);

        action(server, mailbox);
      });
    }

    public static void TestMessage(Action<ImapPseudoServer2, ImapMessageInfo> action)
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        server.EnqueueTaggedResponse("UID FETCH 1 (UID)\r\n",
                                     "* FETCH 1 (UID 1)\r\n" +
                                     "$tag OK done\r\n");

        var message = mailbox.GetMessageByUid(1L);

        Assert.IsTrue(message.IsAvailable);
        Assert.AreEqual(1L, message.Uid);
        Assert.AreEqual(1L, message.Sequence);
        Assert.IsFalse(message.IsDeleted);
        Assert.IsNotNull(message.Url);
        Assert.AreSame(mailbox, message.Mailbox);

        server.DequeueRequest(); // FETCH

        action(server, message);
      });
    }

    public static void TestMessageClosedMailbox(Action<ImapPseudoServer2, ImapMessageInfo> action)
    {
      TestMessage(delegate(ImapPseudoServer2 server, ImapMessageInfo message) {
        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK done\r\n");

        var mailbox = message.Mailbox;

        mailbox.Close();

        server.DequeueRequest(); // CLOSE

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsFalse(message.IsAvailable);
        Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(message.Mailbox));

        action(server, message);
      });
    }

    public static void TestMessageVanished(Action<ImapPseudoServer2, ImapMessageInfo> action)
    {
      TestMessage(delegate(ImapPseudoServer2 server, ImapMessageInfo message) {
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "* EXPUNGE 1\r\n" +
                                     "$tag OK done\r\n");

        message.Client.Refresh();

        server.DequeueRequest(); // NOOP

        Assert.IsTrue(message.IsDeleted);
        Assert.IsFalse(message.IsAvailable);
        Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(message.Mailbox));

        action(server, message);
      });
    }

    public static void TestMessage(string fetchResponse,
                                   ImapMessageFetchAttributeOptions options,
                                   Action<ImapPseudoServer2, ImapMessageInfo> action)
    {
      TestMessages(fetchResponse, options, delegate(ImapPseudoServer2 server, ImapMessageInfo[] messages) {
        action(server, messages.First());
      });
    }

    public static void TestMessages(string fetchResponse,
                                    ImapMessageFetchAttributeOptions options,
                                    Action<ImapPseudoServer2, ImapMessageInfo[]> action)
    {
      var selectResp = "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.ESearch, ImapCapability.Searchres},
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        server.EnqueueTaggedResponse("SEARCH ALL\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("UID FETCH ",
                                     fetchResponse);

        var messages = mailbox.GetMessages(ImapSearchCriteria.All, options);
        var ret = messages.ToArray();

        server.DequeueRequest(); // SEARCH
        server.DequeueRequest(); // FETCH

        Assert.IsTrue(messages.IsAvailable);
        Assert.AreSame(mailbox, messages.Mailbox);

        action(server, ret);
      });
    }
  }
}

==> Smdn.Net.Imap4.Client-1.05.tar.bz2: Smdn.Net.Imap4.Client-1.05/Smdn.Net.Imap4.Client/Test/Smdn.Net.Imap4.Client/TestUtils.cs <==
using System;
using System.Linq;
using NUnit.Framework;

using Smdn.Net.Imap4.Client.Session;

namespace Smdn.Net.Imap4.Client {
  public static class TestUtils {
    public static void TestAuthenticated(Action<ImapPseudoServer, ImapClient> testAction)
    {
      TestAuthenticated(new ImapCapability[0], testAction);
    }

    public static void TestAuthenticated(ImapCapability[] serverCapabilities, Action<ImapPseudoServer, ImapClient> testAction)
    {
      var capas = new ImapCapabilitySet(serverCapabilities ?? new ImapCapability[0]);

      try {
        capas.Add(ImapCapability.Imap4Rev1);
      }
      catch {
        // ignore exception
      }

      using (var server = new ImapPseudoServer()) {
        server.Start();

        // greeting
        server.SetGreetingResponse("* OK ImapPseudoServer ready\r\n");
        // CAPABILITY
        server.EnqueueTaggedResponse("* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");
        // LOGIN
        server.EnqueueTaggedResponse("$tag OK done\r\n");
        // CAPABILITY
        server.EnqueueTaggedResponse(string.Format("* CAPABILITY {0}\r\n", string.Join(" ", capas.ToStringArray())) +
                                     "$tag OK done\r\n");

        using (var client = new ImapClient(new Uri(string.Format("imap://user@{0}/", server.HostPort)))) {
          client.Timeout = 5000;
          client.Profile.AllowInsecureLogin = true;
          client.Profile.UsingSaslMechanisms = null;
          client.Profile.UseTlsIfAvailable = false;

          client.Connect("pass");

          StringAssert.EndsWith("CAPABILITY\r\n", server.DequeueRequest());
          StringAssert.EndsWith("LOGIN user pass\r\n", server.DequeueRequest());
          StringAssert.EndsWith("CAPABILITY\r\n", server.DequeueRequest());

          testAction(server, client);
        }
      }
    }

    public static void TestOpenedMailbox(string mailboxName, Action<ImapPseudoServer, ImapOpenedMailboxInfo> action)
    {
      TestOpenedMailbox(null, mailboxName, "$tag OK [READ-WRITE] done\r\n", action);
    }

    public static void TestOpenedMailbox(string mailboxName, string selectResponse, Action<ImapPseudoServer, ImapOpenedMailboxInfo> action)
    {
      TestOpenedMailbox(null, mailboxName, selectResponse, action);
    }

    public static void TestOpenedMailbox(ImapCapability[] capabilities, string mailboxName, string selectResponse, Action<ImapPseudoServer, ImapOpenedMailboxInfo> action)
    {
      TestUtils.TestAuthenticated(capabilities, delegate(ImapPseudoServer server, ImapClient client) {
        // LIST
        server.EnqueueTaggedResponse(string.Format("* LIST () \"\" {0}\r\n", mailboxName) +
                                     "$tag OK done\r\n");
        // SELECT
        server.EnqueueTaggedResponse(selectResponse);

        using (var opened = client.OpenMailbox(mailboxName)) {
          server.DequeueRequest(); // LIST
          server.DequeueRequest(); // SELECT

          try {
            Assert.IsTrue(opened.IsOpen);
            Assert.AreEqual(opened.IsReadOnly, selectResponse.Contains("OK [READ-ONLY]"));
            Assert.AreEqual(mailboxName, opened.FullName);
            Assert.IsNotNull(opened.ApplicableFlags);
            Assert.IsNotNull(opened.PermanentFlags);

            action(server, opened);
          }
          finally {
            // CLOSE
            if (client.IsConnected && opened.IsOpen)
              server.EnqueueTaggedResponse("$tag OK done\r\n");
          }
        }
      });
    }

    public static void TestClosedMailbox(string mailboxName, Action<ImapPseudoServer, ImapOpenedMailboxInfo> action)
    {
      TestOpenedMailbox(mailboxName, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        // CLOSE
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        mailbox.Close();

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsNull(mailbox.Client.OpenedMailbox);

        Assert.That(server.DequeueRequest(), Is.StringEnding("CLOSE\r\n"));

        action(server, mailbox);
      });
    }
  }

  public static class TestUtils2 {
    public static void TestAuthenticated(Action<ImapPseudoServer2, ImapClient> testAction)
    {
      TestAuthenticated(new ImapCapability[0],
                        testAction);
    }

    public static void TestAuthenticated(ImapCapability[] serverCapabilities,
                                         Action<ImapPseudoServer2, ImapClient> testAction)
    {
      var capas = new ImapCapabilitySet(serverCapabilities ?? new ImapCapability[0]);

      try {
        capas.Add(ImapCapability.Imap4Rev1);
      }
      catch {
        // ignore exception
      }

      using (var server = ImapPseudoServer2.StartNew()) {
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     string.Format("* CAPABILITY {0}\r\n", string.Join(" ", capas.ToStringArray())) +
                                     "$tag OK done\r\n");

        using (var client = new ImapClient(new Uri(string.Format("imap://user@{0}/", server.HostPort)))) {
          client.Timeout = 5000;
          client.Profile.AllowInsecureLogin = true;
          client.Profile.UsingSaslMechanisms = null;
          client.Profile.UseTlsIfAvailable = false;

          client.Connect("pass");

          server.DequeueRequest(); // CAPABILITY
          server.DequeueRequest(); // LOGIN
          server.DequeueRequest(); // CAPABILITY

          testAction(server, client);
        }
      }
    }

    public static void TestDisconnected(Action<ImapPseudoServer2, ImapClient> testAction)
    {
      TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n",
                                     true); // stop server

        client.Refresh();

        server.DequeueRequest();

        testAction(server, client);
      });
    }

    public static void TestLoggedOut(Action<ImapPseudoServer2, ImapClient> testAction)
    {
      TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        server.EnqueueTaggedResponse("LOGOUT\r\n",
                                     "$tag OK logged out\r\n");

        client.Logout();

        server.DequeueRequest();

        testAction(server, client);
      });
    }

    public static void TestMailbox(string name, Action<ImapPseudoServer2, ImapMailboxInfo> action)
    {
      TestMailbox(null, null, ".", name, action);
    }

    public static void TestMailbox(string hierarchyDelimiter, string name, Action<ImapPseudoServer2, ImapMailboxInfo> action)
    {
      TestMailbox(null, null, hierarchyDelimiter, name, action);
    }

    public static void TestMailbox(ImapCapability[] capabilities,
                                   ImapMailboxFlag[] flags,
                                   string name,
                                   Action<ImapPseudoServer2, ImapMailboxInfo> action)
    {
      TestMailbox(capabilities, flags, ".", name, action);
    }

    public static void TestMailbox(ImapCapability[] capabilities,
                                   ImapMailboxFlag[] flags,
                                   string hierarchyDelimiter,
                                   string name,
                                   Action<ImapPseudoServer2, ImapMailboxInfo> action)
    {
      capabilities = capabilities ?? new ImapCapability[0];

      TestAuthenticated(capabilities, delegate(ImapPseudoServer2 server, ImapClient client) {
        var joinedFlags = string.Join(" ", Array.ConvertAll(flags ?? new ImapMailboxFlag[0], delegate(ImapMailboxFlag f) {
          return f.ToString();
        }).ToArray());

        server.EnqueueTaggedResponse("LIST ",
                                     string.Format("* LIST ({0}) \"{1}\" {2}\r\n", joinedFlags, hierarchyDelimiter, name) +
                                     "$tag OK done\r\n");

        var mailbox = client.GetMailbox(name);

        Assert.IsNotNull(mailbox);
        Assert.AreEqual(name, mailbox.FullName);
        Assert.AreEqual(hierarchyDelimiter, mailbox.MailboxSeparator);

        var requested = server.DequeueRequest();

        if (Array.Exists(capabilities, delegate(ImapCapability capa) { return capa == ImapCapability.ListExtended; })) {
          Assert.That(requested, Is.StringContaining("LIST ("));
          Assert.That(requested, Is.StringContaining(string.Format(") \"\" {0}", name)));
        }
        else {
          Assert.That(requested, Is.StringEnding(string.Format("LIST \"\" {0}\r\n", name)));
        }

        action(server, mailbox);
      });
    }

    public static void TestOpenedMailbox(string mailboxName,
                                         Action<ImapPseudoServer2, ImapOpenedMailboxInfo> action)
    {
      TestOpenedMailbox(null,
                        mailboxName,
                        "$tag OK [READ-WRITE] done\r\n",
                        action);
    }

    public static void TestOpenedMailbox(string mailboxName,
                                         string selectResponse,
                                         Action<ImapPseudoServer2, ImapOpenedMailboxInfo> action)
    {
      TestOpenedMailbox(null,
                        mailboxName,
                        selectResponse,
                        action);
    }

    public static void TestOpenedMailbox(ImapCapability[] capabilities,
                                         string mailboxName,
                                         string selectResponse,
                                         Action<ImapPseudoServer2, ImapOpenedMailboxInfo> action)
    {
      TestAuthenticated(capabilities,
                        delegate(ImapPseudoServer2 server, ImapClient client) {
        server.EnqueueTaggedResponse("LIST \"\"",
                                     string.Format("* LIST () \"\" {0}\r\n", mailboxName) +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("SELECT " + mailboxName,
                                     selectResponse);

        using (var opened = client.OpenMailbox(mailboxName)) {
          server.DequeueRequest(); // LIST
          server.DequeueRequest(); // SELECT

          try {
            Assert.IsTrue(opened.IsOpen);
            Assert.AreEqual(opened.IsReadOnly, selectResponse.Contains("OK [READ-ONLY]"));
            Assert.AreEqual(mailboxName, opened.FullName);
            Assert.IsNotNull(opened.ApplicableFlags);
            Assert.IsNotNull(opened.PermanentFlags);

            action(server, opened);
          }
          finally {
            // CLOSE
            if (client.IsConnected && opened.IsOpen)
              server.EnqueueTaggedResponse("CLOSE\r\n",
                                           "$tag OK done\r\n");

          }
        }
      });
    }

    public static void TestClosedMailbox(string mailboxName,
                                         Action<ImapPseudoServer2, ImapOpenedMailboxInfo> action)
    {
      TestOpenedMailbox(mailboxName,
                        delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK done\r\n");

        mailbox.Close();

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsNull(mailbox.Client.OpenedMailbox);

        action(server, mailbox);
      });
    }

    public static void TestMessage(Action<ImapPseudoServer2, ImapMessageInfo> action)
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        server.EnqueueTaggedResponse("UID FETCH 1 (UID)\r\n",
                                     "* FETCH 1 (UID 1)\r\n" +
                                     "$tag OK done\r\n");

        var message = mailbox.GetMessageByUid(1L);

        Assert.IsTrue(message.IsAvailable);
        Assert.AreEqual(1L, message.Uid);
        Assert.AreEqual(1L, message.Sequence);
        Assert.IsFalse(message.IsDeleted);
        Assert.IsNotNull(message.Url);
        Assert.AreSame(mailbox, message.Mailbox);

        server.DequeueRequest(); // FETCH

        action(server, message);
      });
    }

    public static void TestMessageClosedMailbox(Action<ImapPseudoServer2, ImapMessageInfo> action)
    {
      TestMessage(delegate(ImapPseudoServer2 server, ImapMessageInfo message) {
        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK done\r\n");

        var mailbox = message.Mailbox;

        mailbox.Close();

        server.DequeueRequest(); // CLOSE

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsFalse(message.IsAvailable);
        Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(message.Mailbox));

        action(server, message);
      });
    }

    public static void TestMessageVanished(Action<ImapPseudoServer2, ImapMessageInfo> action)
    {
      TestMessage(delegate(ImapPseudoServer2 server, ImapMessageInfo message) {
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "* EXPUNGE 1\r\n" +
                                     "$tag OK done\r\n");

        message.Client.Refresh();

        server.DequeueRequest(); // NOOP

        Assert.IsTrue(message.IsDeleted);
        Assert.IsFalse(message.IsAvailable);
        Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(message.Mailbox));

        action(server, message);
      });
    }

    public static void TestMessage(string fetchResponse,
                                   ImapMessageFetchAttributeOptions options,
                                   Action<ImapPseudoServer2, ImapMessageInfo> action)
    {
      TestMessages(fetchResponse, options, delegate(ImapPseudoServer2 server, ImapMessageInfo[] messages) {
        action(server, messages.First());
      });
    }

    public static void TestMessages(string fetchResponse,
                                    ImapMessageFetchAttributeOptions options,
                                    Action<ImapPseudoServer2, ImapMessageInfo[]> action)
    {
      var selectResp = "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.ESearch, ImapCapability.Searchres},
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        server.EnqueueTaggedResponse("SEARCH ALL\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("UID FETCH ",
                                     fetchResponse);

        var messages = mailbox.GetMessages(ImapSearchCriteria.All, options);
        var ret = messages.ToArray();

        server.DequeueRequest(); // SEARCH
        server.DequeueRequest(); // FETCH

        Assert.IsTrue(messages.IsAvailable);
        Assert.AreSame(mailbox, messages.Mailbox);

        action(server, ret);
      });
    }
  }
}
